import math
from typing import Any, Generic, Iterable, List, Optional, Tuple, TypeVar

import numpy as np
import pandas as pd
import pyarrow as pa

from triad.utils.convert import to_size

from .iter import slice_iterable

T = TypeVar("T")


class BatchReslicer(Generic[T]):
    """Reslice batch streams with row or/and size limit

    :param row_limit: max row for each slice, defaults to None
    :param size_limit: max byte size for each slice, defaults to None

    :raises AssertionError: if `size_limit` is not None but `sizer` is None
    """

    def __init__(
        self,
        row_limit: Optional[int] = None,
        size_limit: Any = None,
    ) -> None:
        if row_limit is None:
            self._row_limit = 0
        else:
            self._row_limit = row_limit
        if size_limit is None:
            self._size_limit = 0
        else:
            self._size_limit = to_size(str(size_limit))

    def get_rows_and_size(self, batch: T) -> Tuple[int, int]:
        """Get the number of rows and byte size of a batch

        :param batch: the batch object
        :return: the number of rows and byte size of the batch
        """
        raise NotImplementedError  # pragma: no cover

    def take(self, batch: T, start: int, length: int) -> T:
        """Take a slice of the batch

        :param batch: the batch object
        :param start: the start row index
        :param length: the number of rows to take

        :return: a slice of the batch
        """
        raise NotImplementedError  # pragma: no cover

    def concat(self, batches: List[T]) -> T:
        """Concatenate a list of batches into one batch

        :param batches: the list of batches
        :return: the concatenated batch
        """
        raise NotImplementedError  # pragma: no cover

    def reslice(self, batches: Iterable[T]) -> Iterable[T]:  # noqa: C901, A003
        """Reslice the batch stream into new batches constrained by the row or size limit

        :param batches: the batch stream

        :yield: an iterable of batches of the same type with the constraints
        """
        if self._row_limit <= 0 and self._size_limit <= 0:
            for batch in batches:
                batch_rows, _ = self.get_rows_and_size(batch)
                if batch_rows > 0:
                    yield batch
            return
        cache: List[T] = []
        total_rows, total_size = 0, 0
        cache_rows = 0
        for batch in batches:
            batch_rows, batch_size = self.get_rows_and_size(batch)
            if batch_rows == 0:
                continue
            total_rows += batch_rows
            total_size += batch_size
            row_size = total_size / total_rows

            if self._row_limit > 0 and self._size_limit <= 0:
                row_limit = self._row_limit
            elif self._row_limit <= 0 and self._size_limit > 0:
                row_limit = max(math.floor(self._size_limit / row_size), 1)
            else:
                row_limit = min(
                    self._row_limit, max(math.floor(self._size_limit / row_size), 1)
                )

            if cache_rows >= row_limit:  # clean up edge cases
                yield self.concat(cache)
                cache = []
                cache_rows = 0

            if cache_rows + batch_rows < row_limit:
                cache.append(batch)
                cache_rows += batch_rows
            else:
                # here we guarantee initial_rows > 0
                slices, remain = self._slice_rows(
                    batch_rows, row_limit - cache_rows, slice_rows=row_limit
                )
                for i, rg in enumerate(slices):
                    chunk = self.take(batch, rg[0], rg[1])
                    if i == 0:
                        yield self.concat(cache + [chunk])
                        cache = []
                        cache_rows = 0
                    else:
                        yield chunk
                if remain[1] > 0:
                    cache.append(self.take(batch, remain[0], remain[1]))
                    cache_rows += remain[1]
        if len(cache) > 0:
            yield self.concat(cache)

    def _slice_rows(
        self, batch_rows: int, initial_rows: int, slice_rows: int
    ) -> Tuple[List[Tuple[int, int]], Tuple[int, int]]:
        start = 0
        if initial_rows >= batch_rows:
            return [(0, batch_rows)], (0, 0)
        slices = [(0, initial_rows)]
        start = initial_rows
        while True:
            if batch_rows - start < slice_rows:
                return slices, (start, batch_rows - start)
            slices.append((start, slice_rows))
            start += slice_rows


class PandasBatchReslicer(BatchReslicer[pd.DataFrame]):
    def get_rows_and_size(self, batch: pd.DataFrame) -> Tuple[int, int]:
        return batch.shape[0], batch.memory_usage(deep=True).sum()

    def take(self, batch: pd.DataFrame, start: int, length: int) -> pd.DataFrame:
        if start == 0 and length == batch.shape[0]:
            return batch
        return batch.iloc[start : start + length]

    def concat(self, batches: List[pd.DataFrame]) -> pd.DataFrame:
        if len(batches) == 1:
            return batches[0]
        return pd.concat(batches)


class ArrowTableBatchReslicer(BatchReslicer[pa.Table]):
    def get_rows_and_size(self, batch: pa.Table) -> Tuple[int, int]:
        return batch.num_rows, batch.nbytes

    def take(self, batch: pa.Table, start: int, length: int) -> pa.Table:
        if start == 0 and length == batch.num_rows:
            return batch
        return batch.slice(start, length)

    def concat(self, batches: List[pa.Table]) -> pa.Table:
        if len(batches) == 1:
            return batches[0]
        return pa.concat_tables(batches)


class NumpyArrayBatchReslicer(BatchReslicer[np.ndarray]):
    def get_rows_and_size(self, batch: np.ndarray) -> Tuple[int, int]:
        return batch.shape[0], batch.nbytes

    def take(self, batch: np.ndarray, start: int, length: int) -> np.ndarray:
        if start == 0 and length == batch.shape[0]:
            return batch
        return batch[start : start + length]

    def concat(self, batches: List[np.ndarray]) -> np.ndarray:
        if len(batches) == 1:
            return batches[0]
        return np.concatenate(batches, axis=0)


class SortedBatchReslicer(Generic[T]):
    """Reslice batch streams (that are alredy sorted by keys) by keys.

    :param keys: group keys to reslice by
    """

    def __init__(
        self,
        keys: List[str],
    ) -> None:
        self._keys = keys
        self._last_row: Optional[np.ndarray] = None

    def take(self, batch: T, start: int, length: int) -> T:
        """Take a slice of the batch

        :param batch: the batch object
        :param start: the start row index
        :param length: the number of rows to take

        :return: a slice of the batch
        """
        raise NotImplementedError  # pragma: no cover

    def concat(self, batches: List[T]) -> T:
        """Concatenate a list of batches into one batch

        :param batches: the list of batches
        :return: the concatenated batch
        """
        raise NotImplementedError  # pragma: no cover

    def get_keys_ndarray(self, batch: T, keys: List[str]) -> np.ndarray:
        """Get the keys as a numpy array

        :param batch: the batch object
        :param keys: the keys to get

        :return: the keys as a numpy array
        """
        raise NotImplementedError  # pragma: no cover

    def get_batch_length(self, batch: T) -> int:
        """Get the number of rows in the batch

        :param batch: the batch object

        :return: the number of rows in the batch
        """
        raise NotImplementedError  # pragma: no cover

    def reslice(
        self, batches: Iterable[T]
    ) -> Iterable[Iterable[T]]:  # noqa: C901, A003
        """Reslice the batch stream into a stream of iterable of batches of the
        same keys

        :param batches: the batch stream

        :yield: an iterable of iterable of batches containing same keys
        """

        def slicer(
            n: int, current: Tuple[bool, T], last: Optional[Tuple[bool, T]]
        ) -> bool:
            return current[0]

        def get_slices() -> Iterable[Tuple[bool, T]]:
            for batch in batches:
                if self.get_batch_length(batch) > 0:
                    yield from self._reslice_single(batch)

        def transform(data: Iterable[Tuple[bool, T]]) -> Iterable[T]:
            for _, batch in data:
                yield batch

        for res in slice_iterable(get_slices(), slicer):
            yield transform(res)

    def reslice_and_merge(
        self, batches: Iterable[T]
    ) -> Iterable[T]:  # noqa: C901, A003
        """Reslice the batch stream into new batches, each containing the same keys

        :param batches: the batch stream

        :yield: an iterable of batches, each containing the same keys
        """

        cache: Optional[T] = None

        for batch in batches:
            if self.get_batch_length(batch) > 0:
                for diff, sub in self._reslice_single(batch):
                    if not diff:
                        cache = self.concat([cache, sub])  # type: ignore
                    else:
                        if cache is not None:
                            yield cache
                        cache = sub

        if cache is not None:
            yield cache

    def _reslice_single(self, batch: T) -> Iterable[Tuple[bool, T]]:
        a = self.get_keys_ndarray(batch, self._keys)
        b = np.roll(a, 1, axis=0)
        diff = self._diff(a, b)
        if self._last_row is not None:
            diff_from_last: bool = self._diff(a[0:1], self._last_row)[0]  # type: ignore
        else:
            diff_from_last = True
        self._last_row = a[-1:]
        points = np.where(diff)[0].tolist() + [a.shape[0]]
        if len(points) == 1:
            yield diff_from_last, batch
        else:
            for i in range(len(points) - 1):
                new_start = diff_from_last if i == 0 else True
                yield new_start, self.take(batch, points[i], points[i + 1] - points[i])

    def _diff(self, a: np.ndarray, b: np.ndarray) -> bool:
        return ((a == b) | ((a != a) & (b != b))).sum(axis=1) < len(self._keys)


class PandasSortedBatchReslicer(SortedBatchReslicer[pd.DataFrame]):
    def get_keys_ndarray(self, batch: pd.DataFrame, keys: List[str]) -> np.ndarray:
        return batch[keys].to_numpy()

    def get_batch_length(self, batch: pd.DataFrame) -> int:
        return batch.shape[0]

    def take(self, batch: pd.DataFrame, start: int, length: int) -> pd.DataFrame:
        return batch.iloc[start : start + length]

    def concat(self, batches: List[pd.DataFrame]) -> pd.DataFrame:
        return pd.concat(batches)


class ArrowTableSortedBatchReslicer(SortedBatchReslicer[pa.Table]):
    def get_keys_ndarray(self, batch: pa.Table, keys: List[str]) -> np.ndarray:
        return batch.select(keys).to_pandas().to_numpy()

    def get_batch_length(self, batch: pa.Table) -> int:
        return batch.num_rows

    def take(self, batch: pa.Table, start: int, length: int) -> pa.Table:
        return batch.slice(start, length)

    def concat(self, batches: List[pa.Table]) -> pa.Table:
        return pa.concat_tables(batches)



from collections import defaultdict
from typing import Any, Dict, Iterable, List, Optional

from .assertion import assert_or_throw
from .string import validate_triad_var_name


def normalize_names(names: List[Any]) -> Dict[Any, str]:
    """Normalize dataframe column names to follow Fugue column naming
    rules. It only operates on names that are not valid to Fugue.

    It tries to minimize the changes to the original name. Special characters
    will be converted to ``_``, but if this does not provide a valid and
    unique column name, more transformation will be done.

    .. note::

        This is a temporary solution before :class:`~.triad.collections.schema.Schema`
        can take arbitrary names

    .. admonition:: Examples

        * ``[0,1]`` => ``{0:"_0", 1:"_1"}``
        * ``["1a","2b"]`` => ``{"1a":"_1a", "2b":"_2b"}``
        * ``["*a","-a"]`` => ``{"*a":"_a", "-a":"_a_1"}``

    :param names: the columns names of a dataframe
    :return: the rename operations as a dict, key is the original column
        name, value is the new valid name.
    """
    assert_or_throw(len(names) > 0, ValueError("names is empty"))
    assert_or_throw(
        len(set(names)) == len(names), ValueError(f"duplicated names found in {names}")
    )
    dup_ct = defaultdict(int)
    dup_ct[""] = 1
    result: Dict[Any, str] = {}
    _names: List[str] = []
    for _name in names:
        if isinstance(_name, str) and validate_triad_var_name(_name):
            dup_ct[_name] += 1
        else:
            _names.append(_name)
    for _name in _names:
        name = None if _name is None else str(_name)
        nn = _normalize_name(name)
        if dup_ct[nn] > 0:
            while dup_ct[nn] > 0:
                orig_nn = nn
                nn = nn + "_" + str(dup_ct[nn])
            dup_ct[orig_nn] += 1
        dup_ct[nn] += 1
        if not isinstance(_name, str) or _name != nn:
            result[_name] = nn
    return result


def _normalize_name(name: Optional[str]) -> str:
    if name is None:
        return ""
    if validate_triad_var_name(name) and not all(x == "_" for x in name):
        return name
    name = name.strip()
    if name == "":
        return ""
    name = "".join(_normalize_chars(name))
    if name[0].isdigit():
        name = "_" + name
    if validate_triad_var_name(name) and not all(x == "_" for x in name):
        return name
    return ""


def _normalize_chars(name: str) -> Iterable[str]:
    for c in name:
        i = ord(c)
        if i < len(_VALID_CHARS) and _VALID_CHARS[i]:
            yield c
        else:
            yield "_"


def _get_valid_signs():
    signs = [False] * 128
    for i in range(len(signs)):
        c = chr(i)
        if c.isalpha() or c.isdigit() or c == "_":
            signs[i] = True
    return signs


_VALID_CHARS = _get_valid_signs()



from typing import Iterable, List, Tuple

from triad.constants import TRIAD_VAR_QUOTE

from .assertion import assert_or_throw
from .string import validate_triad_var_name


def unquote_name(name: str, quote: str = TRIAD_VAR_QUOTE) -> str:
    """If the input is quoted, then get the inner string,
    otherwise do nothing.

    :param name: the name string
    :param quote: the quote char, defaults to `
    :return: the value without `
    """
    if validate_triad_var_name(name):
        return name
    if len(name) >= 2 and name[0] == name[-1] == quote:
        return name[1:-1].replace(quote + quote, quote)
    name = name.strip()
    assert_or_throw(
        len(name) > 0,
        ValueError("empty string is invalid"),
    )
    return name


def quote_name(name: str, quote: str = TRIAD_VAR_QUOTE) -> str:
    """Add quote ` for strings that are not a valid triad var name.

    :param name: the name string
    :param quote: the quote char, defaults to `
    :return: the quoted(if necessary) string
    """
    if validate_triad_var_name(name):
        return name
    return quote + name.replace(quote, quote + quote) + quote


def move_to_unquoted(expr: str, p: int, quote=TRIAD_VAR_QUOTE) -> int:
    """When ``p`` is on a quote, find the position next to the end of
    the quoted part

    :param expr: the original string
    :param p: the current position of ``expr``, and it should be a quote
    :param quote: the quote character
    :raises SyntaxError: if there is an open quote detected
    :return: the position next to the end of the quoted part
    """
    e = p + 1
    le = len(expr)
    while e < le:
        if expr[e] == quote:
            if e + 1 < le and expr[e + 1] == quote:  # escape (``)
                e = e + 2
            else:
                return e + 1
        else:
            e += 1
    raise SyntaxError(f"{expr} contains open quote {quote}")


def split_quoted_string(
    s: str, quote=TRIAD_VAR_QUOTE
) -> Iterable[Tuple[bool, int, int]]:
    """Split ``s`` to a sequence of quoted and unquoted parts.

    :param s: the original string
    :param quote: the quote character

    :yield: the tuple in the format of ``is_quoted, start, end``
    """
    b, e = 0, 0
    le = len(s)
    while e < le:
        if s[e] == quote:
            if e > b:
                yield False, b, e
            b = e
            e = move_to_unquoted(s, e, quote=quote)
            yield True, b, e
            b = e
        else:
            e += 1
    if b < le:
        yield False, b, le


def safe_search_out_of_quote(
    s: str, chars: str, quote=TRIAD_VAR_QUOTE
) -> Iterable[Tuple[int, str]]:
    """Search for chars out of the quoted parts

    :param s: the original string
    :param chars: the charaters to find
    :param quote: the quote character
    :yield: the tuple in format of ``position, char``
    """
    for rg in split_quoted_string(s, quote=quote):
        if not rg[0]:
            for i in range(rg[1], rg[2]):
                if s[i] in chars:
                    yield i, s[i]


def safe_replace_out_of_quote(
    s: str, find: str, replace: str, quote=TRIAD_VAR_QUOTE
) -> str:
    """Replace strings out of the quoted part

    :param s: the original string
    :param find: the string to find
    :param replace: the string used to replace
    :param quote: the quote character

    :return: the string with the replacements
    """
    return "".join(
        s[rg[1] : rg[2]] if rg[0] else s[rg[1] : rg[2]].replace(find, replace)
        for rg in split_quoted_string(s, quote=quote)
    )


def safe_split_out_of_quote(
    s: str,
    sep_chars: str,
    max_split: int = -1,
    quote: str = TRIAD_VAR_QUOTE,
) -> List[str]:
    b = 0
    if max_split == 0 or len(s) == 0:
        return [s]
    res: List[str] = []
    for p, _ in safe_search_out_of_quote(s, sep_chars, quote=quote):
        res.append(s[b:p])
        b = p + 1
        if len(res) == max_split:
            break
    res.append(s[b:])
    return res


def safe_split_and_unquote(
    s: str,
    sep_char: str = ",",
    quote: str = TRIAD_VAR_QUOTE,
    on_unquoted_empty: str = "keep",
) -> List[str]:
    """Split the string and unquote every part

    .. admonition:: Examples

        ``" a , ` b ` , c "`` => ``["a", " b ","c"]``

    :param s: the original string
    :param sep_char: the split character, defaults to ","
    :param quote: the quote character
    :param on_unquoted_empty: can be ``keep``, ``ignore`` or
        ``throw``, defaults to "keep"
    :raises ValueError: if there are empty but unquoted parts and
        ``on_unquoted_empty`` is ``throw``
    :return: the unquoted parts.
    """
    res: List[str] = []
    for _s in safe_split_out_of_quote(s, sep_char, quote=quote):
        s = _s.strip()
        if s == "":
            if on_unquoted_empty == "keep":
                res.append(s)
            elif on_unquoted_empty == "ignore":
                continue
            else:
                raise ValueError(f"empty is not allowed in {s}")
        else:
            res.append(unquote_name(s, quote=quote))
    return res



class NoneArgumentError(ValueError):
    """Exception on None argument"""

    def __init__(self, message: str):
        super().__init__(message)


class InvalidOperationError(Exception):
    """Exception on invalid operations"""

    def __init__(self, message: str = ""):
        super().__init__(message)



import os

from setuptools import find_packages, setup

from triad_version import __version__

with open("README.md") as f:
    LONG_DESCRIPTION = f.read()


def get_version() -> str:
    tag = os.environ.get("RELEASE_TAG", "")
    if "dev" in tag.split(".")[-1]:
        return tag
    if tag != "":
        assert tag == __version__, "release tag and version mismatch"
    return __version__


setup(
    name="triad",
    version=get_version(),
    packages=find_packages(include=["triad*"]),
    package_data={"triad": ["py.typed"]},
    description="A collection of python utils for Fugue projects",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    license="Apache-2.0",
    author="The Fugue Development Team",
    author_email="hello@fugue.ai",
    keywords="fugue utilities",
    url="http://github.com/fugue-project/triad",
    install_requires=[
        "numpy",
        "pandas>=1.3.5",
        "six",
        "pyarrow>=6.0.1",
        "fsspec",
        "fs",  # TODO: remove this hard dependency
    ],
    extras_require={"ciso8601": ["ciso8601"]},
    classifiers=[
        # "3 - Alpha", "4 - Beta" or "5 - Production/Stable"
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3 :: Only",
    ],
    python_requires=">=3.8",
)



import os
import re
import tempfile
import zipfile
from contextlib import contextmanager
from pathlib import Path, PurePosixPath
from typing import Any, Iterator, List, Tuple

import fsspec
import fsspec.core as fc
from fsspec import AbstractFileSystem
from fsspec.implementations.local import LocalFileSystem

_SCHEME_PREFIX = re.compile(r"^[a-zA-Z0-9\-_]+:")


@contextmanager
def chdir(path: str) -> Iterator[None]:
    """Change the current working directory to the given path

    :param path: the path to change to

    .. admonition:: Examples

        .. code-block:: python

            from fugue_ml.utils.io import chdir

            with chdir("/tmp"):
                # do something
    """
    op = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(op)


def url_to_fs(path: str, **kwargs: Any) -> Tuple[AbstractFileSystem, str]:
    """A wrapper of ``fsspec.core.url_to_fs``

    :param path: the path to be used
    :param kwargs: additional arguments to ``fsspec.core.url_to_fs``
    :return: the file system and the path
    """
    if path.startswith("file://"):
        path = path[7:]
    return fc.url_to_fs(path, **kwargs)


def exists(path: str) -> bool:
    """Check if a file or a directory exists

    :param path: the path to check
    :return: whether the path (resource) exists
    """
    fs, path = url_to_fs(path)
    return fs.exists(path)


def isdir(path: str) -> bool:
    """Check if a path is a directory

    :param path: the path to check
    :return: whether the path is a directory
    """
    fs, path = url_to_fs(path)
    return fs.isdir(path)


def isfile(path: str) -> bool:
    """Check if a path is a file

    :param path: the path to check
    :return: whether the path is a file
    """
    fs, path = url_to_fs(path)
    return fs.isfile(path)


def abs_path(path: str) -> str:
    """Get the absolute path of a path

    :param path: the path to check
    :return: the absolute path
    """
    p, _path = fc.split_protocol(path)
    if p is None or p == "file":  # local path
        # Path doesn't work with windows
        return os.path.abspath(_path)
    return path


def touch(path: str, auto_mkdir: bool = False) -> None:
    """Create an empty file or update the timestamp of the file

    :param path: the file path
    :param makedirs: if True, create the directory if not exists,
        defaults to False
    """
    fs, _path = url_to_fs(path)
    if auto_mkdir:
        fs.makedirs(fs._parent(_path), exist_ok=True)
    fs.touch(_path, truncate=True)


def rm(path: str, recursive: bool = False) -> None:
    """Remove a file or a directory

    :param path: the path to remove
    :param recursive: if True and the path is directory,
        remove the directory recursively, defaults to False
    """
    fs, path = url_to_fs(path)
    fs.rm(path, recursive=recursive)


def makedirs(path: str, exist_ok: bool = False) -> str:
    """Create a directory

    :param path: the directory path
    :param exist_ok: if True, do not raise error if the directory exists,
        defaults to False

    :return: the absolute directory path
    """
    fs, _path = url_to_fs(path)
    fs.makedirs(_path, exist_ok=exist_ok)
    if isinstance(fs, LocalFileSystem):
        return str(Path(_path).resolve())
    return path


def join(base_path: str, *paths: str) -> str:
    """Join paths with the base path

    :param base_path: the base path
    :param paths: the paths to join to the base path
    :return: the joined path
    """
    if len(paths) == 0:
        return base_path
    p, path = fc.split_protocol(base_path)
    if p is None:  # local path
        return str(Path(base_path).joinpath(*paths))
    return p + "://" + str(PurePosixPath(path).joinpath(*paths))


def glob(path: str) -> List[str]:
    """Glob files

    :param path: the path to glob
    :return: the matched files (absolute paths)
    """
    fs, _path = url_to_fs(path)
    return [fs.unstrip_protocol(x) for x in fs.glob(_path)]


def write_text(path: str, contents: str) -> None:
    """Write text to a file. If the directory of the file does not exist, it
    will create the directory first

    :param path: the file path
    :param contents: the text to write
    """
    with fsspec.open(path, "w") as f:
        f.write(contents)


def read_text(path: str) -> str:
    """Read text from a file

    :param path: the file path
    :return: the text
    """
    with fsspec.open(path, "r") as f:
        return f.read()


def write_bytes(path: str, contents: bytes, create_dir: bool = True) -> None:
    """Write bytes to a file. If the directory of the file does not exist, it
    will create the directory first

    :param path: the file path
    :param contents: the bytes to write
    :param create_dir: if True, create the directory if not exists,
        defaults to True
    """
    fs, path = url_to_fs(path)
    with fs.open(path, "wb") as f:
        f.write(contents)


def read_bytes(path: str) -> bytes:
    """Read bytes from a file

    :param path: the file path
    :return: the bytes
    """
    fs, path = url_to_fs(path)
    with fs.open(path, "rb") as f:
        return f.read()


@contextmanager
def zip_temp(fobj: Any) -> Iterator[str]:
    """Zip a temporary directory to a file object.

    :param fobj: the file path or file object

    .. admonition:: Examples

        .. code-block:: python

            from fugue_ml.utils.io import zip_temp
            from io import BytesIO

            bio = BytesIO()
            with zip_temp(bio) as tmpdir:
                # do something with tmpdir (string)
    """
    if isinstance(fobj, str):
        with fsspec.open(fobj, "wb", create_dir=True) as f:
            with zip_temp(f) as tmpdir:
                yield tmpdir
    else:
        with tempfile.TemporaryDirectory() as tmpdirname:
            yield tmpdirname

            with zipfile.ZipFile(
                fobj, "w", zipfile.ZIP_DEFLATED, allowZip64=True
            ) as zf:
                for root, _, filenames in os.walk(tmpdirname):
                    for name in filenames:
                        file_path = os.path.join(root, name)
                        rel_dir = os.path.relpath(root, tmpdirname)
                        rel_name = os.path.normpath(os.path.join(rel_dir, name))
                        zf.write(file_path, rel_name)


@contextmanager
def unzip_to_temp(fobj: Any) -> Iterator[str]:
    """Unzip a file object into a temporary directory.

    :param fobj: the file object

    .. admonition:: Examples

        .. code-block:: python

            from fugue_ml.utils.io import zip_temp
            from io import BytesIO

            bio = BytesIO()
            with zip_temp(bio) as tmpdir:
                # create files in the tmpdir (string)

            with unzip_to_temp(BytesIO(bio.getvalue())) as tmpdir:
                # read files from the tmpdir (string)
    """
    if isinstance(fobj, str):
        with fsspec.open(fobj, "rb") as f:
            with unzip_to_temp(f) as tmpdir:
                yield tmpdir
    else:
        with tempfile.TemporaryDirectory() as tmpdirname:
            with zipfile.ZipFile(fobj, "r") as zip_ref:
                zip_ref.extractall(tmpdirname)

            yield tmpdirname
